//
// bluecms.cpp
// ===========
//
// Blue CMS Static Content-Management-System
//
// Copyright (c) 2012 IDEAL Software GmbH, Neuss, Germany.
//
//                 www.IdealSoftware.com
//
// This file is part of Blue CMS.
//
// Blue CMS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Blue CMS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Blue CMS.  If not, see <http://www.gnu.org/licenses/>.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <string>
#include <vector>
#include <list>
#include <map>
#include <algorithm>
using namespace std;


#ifdef _MSC_VER
	#define PATH_SEPARATOR	"\\"
#else
	#define PATH_SEPARATOR	"/"
#endif

#define VERSION		"v1.0"


// ==============================================================================================
//										class CParameters
// ==============================================================================================
class CParameters
{
public:
	bool	m_bVerbose;
	bool	m_bWarnEmptyMenuLink;
	bool	m_bWarnNoMenuFile;
	bool	m_bWarnUndefinedMetaTag;
	string	m_strInputFile;
	string	m_strTemplatePath;
	string	m_strOutputFile;
	string	m_strMenuFile;
	string	m_strMenuTemplate;
	string	m_strSitePrefix;

	map<string, string>	m_mapContentFiles;

	CParameters()
		:	m_bVerbose(false),
			m_bWarnEmptyMenuLink(true),
			m_bWarnNoMenuFile(true),
			m_bWarnUndefinedMetaTag(true)
	{
	}
};


// ==============================================================================================
//										class CMenuEntry
// ==============================================================================================
class CMenuEntry
{
public:
	string	m_strMenuTitle;
	string	m_strMenuLink;
	int		m_nLevel;		// menu level depth

	CMenuEntry(const string &title, const string &link, int level)
	{
		m_strMenuTitle	= title;
		m_strMenuLink	= link;
		m_nLevel		= level;
	}
};


// ==============================================================================================
//										Defines
// ==============================================================================================
#define HIGHLIGHT_OFFSET		16000	// for menu style
#define HIGHLIGHT_LINK_OFFSET	32000	// for menu style
#define SELECTED_OFFSET			48000	// for menu style

enum EMenuStyle
{
	enMsNormal,
	enMsHighlight,
	enMsHighlightLink,
	enMsSelected,
};


// ==============================================================================================
//										Globals
// ==============================================================================================
CParameters gParameters;
map<string, string> gMetaTags;
vector<CMenuEntry> gMenu;
map<int, string> gMenuTemplate;
FILE *gfhOut;


// ==============================================================================================
// Output error message and terminate program
// ==============================================================================================
void Error(const char *s, ...)
{
   char buffer[512];
   va_list argptr;

   va_start(argptr, s);
   vsprintf(buffer, s, argptr);
   
   va_end(argptr);
   fprintf(stderr, "\n\nError: %s\n", buffer);
   exit(1);
}


// ==============================================================================================
//											Left
// ==============================================================================================
inline string Left(const string &s, size_t count)
{
	return s.substr(0, count);
}


// ==============================================================================================
//											Right
// ==============================================================================================
inline string Right(const string &s, size_t count)
{
	return s.substr(s.length() - count);
}


// ========================================================================
//                            FindReplace
//
// This seems to be non-optimal at the first glance, but it is recursion
// safe, i.e. "a" can be safely replaced with "aa".
// ========================================================================
string FindReplace(const string &source, const string &pattern, const string &replace)
{
	size_t j = 0;
	string newstr;

	while (j + pattern.length() <= source.length())
	{
		if (source.substr(j, pattern.length()) == pattern)
		{
			newstr += replace;
			j += pattern.length();
		}
		else
			newstr += source[j++];
	}

	// append rest of source
	newstr += source.substr(j, source.length() - j);

	return newstr;
}


// ==============================================================================================
//											ExtractFileName
// ==============================================================================================
string ExtractFileName(const string &s)
{
	string::size_type pos = s.rfind(PATH_SEPARATOR);
	if (pos == string::npos)
		return s;	// not found

	return s.substr(pos + 1);
}


// ==============================================================================================
//											ExtractFilePath
// ==============================================================================================
string ExtractFilePath(const string &s)
{
	string::size_type pos = s.rfind(PATH_SEPARATOR);
	if (pos == string::npos)
		return "";	// not found

	return s.substr(0, pos);
}


// ==============================================================================================
//											StripLF
// ==============================================================================================
void StripLF(char *s)
{
   size_t i;

   for (i=strlen(s)-1; i>=0 && (s[i]=='\r' || s[i]=='\n'); i--)
      s[i] = 0;
}


// ==============================================================================================
//											xfgets
//
// Strips Line-Feed characters from the string
// ==============================================================================================
char *xfgets(char *buf, int size, FILE *fh)
{
   char *rc;

   rc = fgets(buf, size, fh);
   StripLF(buf);

   return rc;
}


// ==============================================================================================
//											ReadTextFile()
//
// Reads the specified text file line by line into a list of strings.
// ==============================================================================================
list<string> ReadTextFile(const string &file_name)
{
#ifdef _MSC_VER
	string fname = FindReplace(file_name, "/", "\\");
#else
	string fname = file_name;
#endif

	FILE *fh = fopen(fname.c_str(), "rt");
	if (!fh)
		Error("failed to open file '%s'", fname.c_str());

	list<string> lines;
	char line[4096];
	while (xfgets(line, sizeof(line), fh))
		lines.push_back(line);

	fclose(fh);
	return lines;
}


// ==============================================================================================
//											ReadFile()
//
// Allocates a buffer and reads the given file into this buffer.
//
// The caller must free the buffer!
// ==============================================================================================
char *ReadFile(const string &file_name)
{
#ifdef _MSC_VER
	string fname = FindReplace(file_name, "/", "\\");
#else
	string fname = file_name;
#endif

	struct stat st;
	if (stat(fname.c_str(), &st) != 0)
		Error("stat failed: %s", fname.c_str());

	char *buffer = (char *)malloc(st.st_size + 1);
	FILE *fh = fopen(fname.c_str(), "rb");

	if (!fh)
		Error("Could not open file: %s", fname.c_str());

	off_t did_read = (off_t)fread(buffer, 1, st.st_size, fh);
	if (did_read != st.st_size)
	{
		fclose(fh);
		Error("Read error, file: %s (did_read = %d, st_size = %d)", fname.c_str(), did_read, st.st_size);
	}

	fclose(fh);
	buffer[st.st_size] = '\0';

	return buffer;
}


// ==============================================================================================
//											SampleMetaTags()
// ==============================================================================================
void SampleMetaTags()
{
	char *buffer;

	for (map<string, string>::iterator it = gParameters.m_mapContentFiles.begin(); it != gParameters.m_mapContentFiles.end(); it++)
	{
		buffer = ReadFile(it->second);
		char *p = strstr(buffer, "@[DefTag:");
		while (p)
		{
			p += 9;

			// Get the name of the meta-tag
			string meta_tag;
			while (*p && *p != ':' && *p != ']')
				meta_tag += *p++;

			if (!*p)
				Error("Bad meta-tag definition in file %s", it->second.c_str());
			else if (*p == ':')
			{
				// Get the value of the meta-tag
				string value;
				p++;
				while (*p && *p != ']')
					value += *p++;

				if (!*p)
					Error("Bad meta-tag definition in file %s", it->second.c_str());

				if (gMetaTags.find(meta_tag) != gMetaTags.end())
					Error("Meta-Tag %s already defined", meta_tag.c_str());

				gMetaTags[meta_tag] = value;
			}
			else
				gMetaTags[meta_tag] = "";

			p = strstr(p, "@[DefTag:");
		}

		free(buffer);
	}
}


// ==============================================================================================
//											CheckTagDefined()
// ==============================================================================================
bool CheckTagDefined(const string &tag)
{
	return gMetaTags.find(tag) != gMetaTags.end();
}


// ==============================================================================================
//											CompileMenu()
// ==============================================================================================
void CompileMenu()
{
	// Compile the menu file
	if (gParameters.m_strMenuFile.empty())
		return;

	list<string> menu = ReadTextFile(gParameters.m_strMenuFile);
	int prev_level = -1;

	for (list<string>::iterator it = menu.begin(); it != menu.end(); it++)
	{
		if (!(*it).empty())
		{
			string title, link;
			string::size_type pos = (*it).find("//");
			if (pos != string::npos)
			{
				title = Left(*it, pos);
				link = (*it).substr(pos + 2);
			}
			else
				title = *it;

			// count TABS in front of title, to get the menu-level
			int level = 0;
			while (title[level] == '\t')
				level++;

			// and remove TABS
			if (level > 0)
				title = title.substr(level);

			// semantic check
			if (level > prev_level + 1)
			{
				if (prev_level == -1)
					Error("The first menu entry must be at level 0, no TABS are allowed.");
				else
					Error(	"The level of a submenu entry may only be one level higher than the previous entry.\n",
							"entry: %s", (*it).c_str());
			}
			prev_level = level;

			gMenu.push_back(CMenuEntry(title, link, level));
		}
	}

	// Compile the menu template file
	list<string> menu_tpl = ReadTextFile(gParameters.m_strMenuTemplate);
	for (list<string>::iterator it = menu_tpl.begin(); it != menu_tpl.end(); it++)
	{
		if (!(*it).empty())
		{
			string s = *it;
			string::size_type pos = s.find(":");
			if (pos == string::npos)
				Error("missing ':' in menu template file");

			EMenuStyle style;
			int offset = 0;
			switch (s[0])
			{
			case 'N':
				style = enMsNormal;;
				break;

			case 'H':
				style = enMsHighlight;
				offset = HIGHLIGHT_OFFSET;
				break;

			case 'L':
				style = enMsHighlightLink;
				offset = HIGHLIGHT_LINK_OFFSET;
				break;

			case 'S':
				style = enMsSelected;
				offset = SELECTED_OFFSET;
				break;

			default:
				Error(	"No valid menu-style (N, H, L S) defined in menu template.\n",
						"entry: %s", (*it).c_str());
			}

			s = s.substr(1);
			int level = atoi(Left(s, pos).c_str()) - 1;	// level starts at "1" in template, so subtract "1"
			level += offset;

			string html = s.substr(pos + 1);

			if (gMenuTemplate.find(level) != gMenuTemplate.end())
				Error("Menu level #%d already defined in menu template", level);

			gMenuTemplate[level] = html;
		}
	}
}



// ==============================================================================================
//											OutputMenuEntry()
// ==============================================================================================
void OutputMenuEntry(int i, EMenuStyle style = enMsNormal)
{
	// get the HTML template for the menu level
	int level = gMenu[i].m_nLevel;

	string title = gMenu[i].m_strMenuTitle;
	string link = gMenu[i].m_strMenuLink;

	if (link.empty())
	{
		// for empty links, use the link of any direct successor-entry with a higher level
		int current_level = level;
		int j = i + 1;
		while (j < gMenu.size() && gMenu[j].m_strMenuLink.empty() && gMenu[j].m_nLevel > current_level)
		{
			current_level = gMenu[j].m_nLevel;
			j++;
		}

		if (j < gMenu.size() && gMenu[j].m_nLevel > current_level)
			link = gMenu[j].m_strMenuLink;

		if (link.empty() && gParameters.m_bWarnEmptyMenuLink)
			fprintf(stderr, "Warning: empty menu-link for menu-title '%s'\n", title.c_str());
	}

	if (style == enMsHighlight)
	{
		if (!link.empty())
			style = enMsHighlightLink;
	}

	if (style == enMsHighlight)
		level += HIGHLIGHT_OFFSET;
	else if (style == enMsHighlightLink)
		level += HIGHLIGHT_LINK_OFFSET;
	else if (style == enMsSelected)
		level += SELECTED_OFFSET;

	string html = gMenuTemplate[level];

	html = FindReplace(html, "@[MenuTitle]", title);
	html = FindReplace(html, "@[MenuLink]", link);

	fprintf(gfhOut, "%s\n", html.c_str());
}


// ==============================================================================================
//											OutputMenu()
// ==============================================================================================
void OutputMenu()
{
	if (gParameters.m_strMenuFile.empty() && gParameters.m_bWarnNoMenuFile)
	{
		fprintf(stderr, "Warning: a menu is referenced in input HTML template, but no menu file was specified.\n");
		return;
	}

	// The current output-name specifies the selected menu entry. Find it.
	string out_file = gParameters.m_strOutputFile;
	string suffix = Right(out_file, 3);
	if (suffix == ".de" || suffix == ".en")
		out_file = Left(out_file, out_file.length() - 3);

	int i;
	for (i = 0; i < gMenu.size(); i++)
	{
		if (gMenu[i].m_strMenuLink == out_file)
			break;
	}

	if (i >= gMenu.size())
	{
		// the current output file is not referenced by the menu.
		// output the whole menu unfolded at level 0
		for (i = 0; i < gMenu.size(); i++)
		{
			if (gMenu[i].m_nLevel == 0)
				OutputMenuEntry(i);
		}
	}
	else
	{
		int level = gMenu[i].m_nLevel;

		// find the top-level parent of the selected entry
		// and all entries, which are parents and must be highlighted
		vector<int> highlight_entries;
		int highlight_level = level;
		int direct_parent = -1;
		int n = i;
		if (level > 0)
		{
			while (n >= 0)
			{
				// find all entries, which are parents and must be highlighted
				if (gMenu[n].m_nLevel < highlight_level)
				{
					highlight_level = gMenu[n].m_nLevel;
					highlight_entries.push_back(n);
					if (direct_parent < 0)
						direct_parent = n;
				}

				if (gMenu[n].m_nLevel == 0)
					break;

				n--;
			}
		}

		// output all top-level entries folded up to the parent of the selected entry
		int j;
		for (j = 0; j < n; j++)
		{
			if (gMenu[j].m_nLevel == 0)
				OutputMenuEntry(j);
		}

		// output all entries from the parent of the selected entry up to - but excluding - the selected entry
		if (n < 0)
			n = 0;
		for (j = n; j < i; j++)
		{
			if (gMenu[j].m_nLevel < level)
			{
				if (find(highlight_entries.begin(), highlight_entries.end(), j) != highlight_entries.end())
					OutputMenuEntry(j, enMsHighlight);
				else
					OutputMenuEntry(j);
			}
			else if (j > direct_parent && gMenu[j].m_nLevel == level)
				OutputMenuEntry(j);
		}

		// Output the selected entry
		OutputMenuEntry(j, enMsSelected);
		j++;

		// output the next deeper level of the selected entry
		while (j < gMenu.size() && gMenu[j].m_nLevel > level)
		{
			if (gMenu[j].m_nLevel == level + 1)
				OutputMenuEntry(j);
			j++;
		}

		// output all entries which have the same level or any higher level than the selected entry
		while (j < gMenu.size())
		{
			if (gMenu[j].m_nLevel <= level)
			{
				OutputMenuEntry(j);
				if (gMenu[j].m_nLevel < level)
					level = gMenu[j].m_nLevel;
			}

			j++;
		}
	}
}


// =====================================================================================================
//												Process()
//
// Processes input text and outputs it to gfhOut.
// Scans through the supplied buffer for Renderer-Tags and replaces them.
// =====================================================================================================
bool banner = false;
void Process(const char *buffer)
{
	const char *p = buffer;
	while (*p)
	{
		if (!banner && *p == 0x0a)
		{
			p++;
			if (*p == 0x0d)
				p++;
			fprintf(gfhOut, "\n\n<!-- This web page was created using Blue CMS "VERSION" - The open source static CMS from www.IdealSoftware.com -->\n\n");
			banner = true;
		}
		else if (*p == '@' && *(p + 1) == '[')
		{
			// read token until ']' or ':' is reached
			char token[256];
			char parameter[4096];
			char addon[4096];
			p += 2;
			char *s = token;
			while (*p && *p != ']' && *p != ':')
				*s++ = *p++;

			*s++ = '\0';

			*parameter = '\0';
			if (*p == ':')
			{
				// read the parameter
				p++;
				s = parameter;
				while (*p && *p != ']' && *p != '%')
					*s++ = *p++;

				*s++ = '\0';
			}

			*addon = '\0';
			if (*p == '%')
			{
				// read the add-on text, e.g. java script code
				p++;
				s = addon;
				while (*p && *p != ']')
					*s++ = *p++;

				*s++ = '\0';
			}

			if (*p != ']')
				Error("missing ']' in input HTML template");
			p++;

			// Tag
			if (strcmp(token, "Tag") == 0)
			{
				if (*parameter == '\0')
					Error("missing label for Meta-tag in input HTML template");
				
				if (gMetaTags.find(parameter) == gMetaTags.end())
				{
					if (gParameters.m_bWarnUndefinedMetaTag)
						fprintf(stderr, "Warning: meta-tag '%s' referenced in input HTML template, but is not defined\n", parameter);
				}
				else
					fprintf(gfhOut, "%s", gMetaTags[parameter].c_str());
			}
			// Menu
			else if (strcmp(token, "Menu") == 0)
			{
				OutputMenu();
			}
			// Content
			else if (strcmp(token, "Content") == 0)
			{
				if (*parameter == '\0')
					Error("missing label for Content-tag in input HTML template");
				string content_file = gParameters.m_mapContentFiles[parameter];
				char *content = ReadFile(content_file);
				char *body = strstr(content, "<body>");
				if (!body)
					body = strstr(content, "<BODY>");
				if (!body)
					Error("could not locate HTML <body> tag in content-file %s", content_file.c_str());
				body += 6;
				char *body_end = strstr(content, "</body>");
				if (!body_end)
					body_end = strstr(content, "</BODY>");
				if (!body_end)
					Error("could not locate HTML </body> tag in content-file %s", content_file.c_str());
				*body_end = '\0';
				Process(body);
				free(content);
			}
			// If <ident>
			else if (strcmp(token, "If") == 0)
			{
				if (*parameter == '\0')
					Error("missing ident for IF-tag in input HTML template");

				if (!CheckTagDefined(parameter))
				{
					// condition failed, search matching @[Else] or @[End] tag
					int if_count = 1;
	
					while (if_count > 0)
					{
						// search next renderer tag
						p = strstr(p, "@[");
						if (!p)
							Error("missing @[End] tag for if-tag in input HTML template");

						if (strncmp(p, "@[If:", 5) == 0)
						{
							if_count++;
							p += 5;
						}
						else if (strncmp(p, "@[Else]", 7) == 0)
						{
							if (if_count == 1)
								if_count--;
							p += 7;
						}
						else if (strncmp(p, "@[End]", 6) == 0)
						{
							if_count--;
							p += 6;
						}
						else
							p++;
					}
				}
			}
			// Else
			else if (strcmp(token, "Else") == 0)
			{
				// skip until matching @[End] tag
				int if_count = 1;

				while (if_count > 0)
				{
					// search next renderer tag
					p = strstr(p, "@[");
					if (!p)
						Error("missing @[End] tag for if-tag in input HTML template");

					if (strncmp(p, "@[If:", 5) == 0)
					{
						if_count++;
						p += 5;
					}
					else if (strncmp(p, "@[End]", 6) == 0)
					{
						if_count--;
						p += 6;
					}
					else
						p++;
				}
			}
			// End
			else if (strcmp(token, "End") == 0)
			{
				// do nothing, just overread
			}
			// Include
			else if (strcmp(token, "Include") == 0)
			{
				if (*parameter == '\0')
					Error("missing file name for INCLUDE-tag in input HTML template");

				string include_file = gParameters.m_strTemplatePath + PATH_SEPARATOR + parameter;
				char *buffer = ReadFile(include_file);
				Process(buffer);
				free(buffer);
			}
			// DefTag
			else if (strcmp(token, "DefTag") == 0)
			{
				// do nothing
			}
			// RelativeURL
			else if (strcmp(token, "RelativeURL") == 0)
			{
				fprintf(gfhOut, "/%s", gParameters.m_strOutputFile.c_str());
			}
			// SiteURL
			else if (strcmp(token, "SiteURL") == 0)
			{
				if (gParameters.m_strSitePrefix.empty())
					Error("using 'SiteURL' tag, but the Site-Prefix is undefined (which is a command line parameter)");
				fprintf(gfhOut, "%s/%s", gParameters.m_strSitePrefix.c_str(), gParameters.m_strOutputFile.c_str());
			}
			// Unknown tag
			else
				Error("unrecognized tag: '%s' in input HTML template", token);
		}
		else
			fputc(*p++, gfhOut);
	}
}


// ==============================================================================================
//											GetFileTime()
// ==============================================================================================
time_t GetFileTime(const string &file_name)
{
	struct stat stat_target;

	if (gParameters.m_bVerbose)
		printf("\ttesting file %s\n", file_name.c_str());

	if (stat(file_name.c_str(), &stat_target) != 0)
		return 0;	// could not stat target, so indicate make

	return stat_target.st_mtime;

}


// ==============================================================================================
//										CheckIncludeFiles()
// ==============================================================================================
bool CheckIncludeFiles(time_t target_time, string file_name)
{
	char param[4096];

	char *buffer = ReadFile(file_name);

	const char *p = strstr(buffer, "@[Include:");
	while (p)
	{
		// read the parameter
		p += 10;
		char *s = param;
		while (*p && *p != ']')
			*s++ = *p++;

		*s++ = '\0';

		if (*param == '\0')
			Error("missing file name for INCLUDE-tag in input HTML template");

		string include_file = gParameters.m_strTemplatePath + PATH_SEPARATOR + param;
		time_t source_time = GetFileTime(include_file);
		if (source_time > target_time || CheckIncludeFiles(target_time, include_file))
		{
			free(buffer);
			return true;
		}

		p = strstr(p, "@[Include:");
	}

	free(buffer);
	return false;
}


// ==============================================================================================
//											IsUpToDate()
// ==============================================================================================
bool IsUpToDate()
{
	time_t source_time;

	// check target
	time_t target_time = GetFileTime(gParameters.m_strOutputFile);
	if (target_time == 0)
		return false;		// target does not exist

	// check input template
	// ==> not necessary, the input template is in gParameters.m_mapContentFiles

	// check menu and menu template
	if (!gParameters.m_strMenuFile.empty())
	{
		source_time = GetFileTime(gParameters.m_strMenuFile);
		if (source_time > target_time)
			return false;

		source_time = GetFileTime(gParameters.m_strMenuTemplate);
		if (source_time > target_time)
			return false;
	}

	// Check content files
	for (map<string, string>::iterator it = gParameters.m_mapContentFiles.begin(); it != gParameters.m_mapContentFiles.end(); it++)
	{
		source_time = GetFileTime(it->second);
		if (source_time > target_time)
			return false;
	}

	// Check include files
	if (CheckIncludeFiles(target_time, gParameters.m_strInputFile))
		return false;

	return true;
}


// ==============================================================================================
//											Compile()
// ==============================================================================================
void Compile()
{
	// If all input files are not newer than the output file, quit
	if (IsUpToDate())
	{
		if (gParameters.m_bVerbose)
			printf("file %s is up-to-date, done.\n", gParameters.m_strOutputFile.c_str());
		return;
	}

	fprintf(stderr, "generating file %s ... ", gParameters.m_strOutputFile.c_str());

	// Read all content files and sample Meta-Tag definitions
	SampleMetaTags();

	// Compile the menu
	CompileMenu();

	// Read Input File
	char *buffer = ReadFile(gParameters.m_strInputFile);

	// Parse input and create output
	gfhOut = fopen(gParameters.m_strOutputFile.c_str(), "w");
	if (!gfhOut)
		Error("can not create output file %s", gParameters.m_strOutputFile.c_str());
	Process(buffer);
	fclose(gfhOut);
	free(buffer);
	fprintf(stderr, "done.\n");
}


// ==============================================================================================
//											Banner()
// ==============================================================================================
void Banner()
{
	printf("\nbluecms "VERSION" - Copyright (c) 2012 IDEAL Software GmbH, Neuss, Germany\n");
}


// ==============================================================================================
//											Usage()
// ==============================================================================================
void Usage()
{
	Banner();
	fprintf(stderr, "Usage: bluecms [-v] [-wl] [-wf] [-wt] -i <input file> -o <output file> -m <menu file> -mt <menu template file>\n"
					"\t-c <label>=<content file> -c <label>=<content file>... -c <label>=<content file>\n\n"
					"\t-v  = verbose\n"
					"\t-wl = do not warn, if a menu link is empty\n"
					"\t-wf = do not warn, if a menu is referenced in input HTML template, but no menu file is specified\n"
					"\t-wt = do not warn, if a meta-tag is undefined\n"
					"\n");
   exit(1);
}


#define CHECK_PARAM		i++; \
						if (i > argc) \
							Usage()


// ==============================================================================================
//											main
// ==============================================================================================
int main(int argc, const char **argv)
{
	// Parse Commandline
	for (int i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "-i") == 0)
		{
			CHECK_PARAM;
			gParameters.m_strInputFile = argv[i];
			gParameters.m_strTemplatePath = ExtractFilePath(gParameters.m_strInputFile);

			// Add template to map of content files, so @[DefTag] commands are evaluated
			gParameters.m_mapContentFiles["__template_file__"] = gParameters.m_strInputFile;
		}
		else if (strcmp(argv[i], "-o") == 0)
		{
			CHECK_PARAM;
			gParameters.m_strOutputFile = argv[i];
		}
		else if (strcmp(argv[i], "-m") == 0)
		{
			CHECK_PARAM;
			gParameters.m_strMenuFile = argv[i];
		}
		else if (strcmp(argv[i], "-s") == 0)
		{
			CHECK_PARAM;
			gParameters.m_strSitePrefix = argv[i];
		}
		else if (strcmp(argv[i], "-mt") == 0)
		{
			CHECK_PARAM;
			gParameters.m_strMenuTemplate = argv[i];
		}
		else if (strcmp(argv[i], "-c") == 0)
		{
			CHECK_PARAM;
			char *s = strdup(argv[i]);
			char *p = strchr(s, '=');
			if (!p)
				Usage();
			*p++ = '\0';
			gParameters.m_mapContentFiles[s] = p;
			free(s);
		}
		else if (strcmp(argv[i], "-v") == 0)
		{
			gParameters.m_bVerbose = true;
		}
		else if (strcmp(argv[i], "-wl") == 0)
		{
			gParameters.m_bWarnEmptyMenuLink = false;
		}
		else if (strcmp(argv[i], "-wf") == 0)
		{
			gParameters.m_bWarnNoMenuFile = false;
		}
		else if (strcmp(argv[i], "-wt") == 0)
		{
			gParameters.m_bWarnUndefinedMetaTag = false;
		}
		else
			Usage();
	}

	if (gParameters.m_bVerbose)
		Banner();

	if (gParameters.m_strInputFile.empty() ||
		gParameters.m_strOutputFile.empty())
	{
		Usage();
	}

	if (gParameters.m_strInputFile == gParameters.m_strOutputFile)
		Error("input and output file must be different!");

	if (!gParameters.m_strMenuFile.empty() && gParameters.m_strMenuTemplate.empty())
		Error("If a menu file is specified on the command line, the menu template file MUST also be specified!");

	Compile();

	return 0;
}
